import torch
from torch import nn

from bayne.bounds.crown import alpha_beta_sequential, linear_bounds
from bayne.bounds.util import notnull, add_method


def crown_ibp(model):
    @torch.no_grad()
    def crown_ibp(self: nn.Sequential, lower, upper):
        with notnull(getattr(self, '_pyro_context', None)):
            batch_size = lower.size(0)

            LBs, UBs = self.ibp(lower, upper, pre=True)

            alpha, beta = alpha_beta_sequential(self, LBs, UBs)
            bounds = linear_bounds(self, alpha, beta, batch_size, lower.device)

            return bounds

    add_method(model, 'crown_ibp', crown_ibp)

    return model



import torch
from torch import nn, distributions
import torch.nn.functional as F

from pyro.nn import PyroModule, PyroSample
from pyro.infer.mcmc import HMC, MCMC, NUTS
from pyro.infer import Predictive
import pyro.distributions as dist
import pyro


class PyroBatchLinear(nn.Linear, PyroModule):
    def __init__(self, in_features: int, out_features: int, bias: bool = True, device=None, dtype=None, weight_prior=None, bias_prior=None) -> None:
        # While calling super().__init__() creates the weights and we overwrite them later, it's just easier this way,
        # and we get to inherit from nn.Linear. It shouldn't be too much of an issue considering that you usually
        # only instantiate a model once during execution.
        super(PyroBatchLinear, self).__init__(in_features, out_features, bias, device, dtype)

        if weight_prior is None:
            weight_prior = dist.Normal(torch.as_tensor(0.0, device=device), torch.as_tensor(1.0, device=device)) \
                                            .expand(self.weight.shape) \
                                            .to_event(self.weight.dim())

        if bias and bias_prior is None:
            bias_prior = dist.Normal(torch.as_tensor(0.0, device=device), torch.as_tensor(0.5, device=device)) \
                                            .expand(self.bias.shape) \
                                            .to_event(self.bias.dim())

        self.weight_prior = weight_prior
        self.weight = PyroSample(prior=self.weight_prior)
        if bias:
            self.bias_prior = bias_prior
            self.bias = PyroSample(prior=self.bias_prior)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        bias = self.bias
        weight = self.weight

        if weight.dim() == 2:
            return F.linear(input, weight, bias)
        else:
            if input.dim() == 2:
                input = input.unsqueeze(0).expand(weight.size(0), *input.size())

            return torch.baddbmm(bias.unsqueeze(1), input, weight.transpose(-1, -2))

    def to(self, *args, **kwargs):
        self.dist_to(self.weight_prior, *args, **kwargs)
        if self.bias is not None:
            self.dist_to(self.bias_prior, *args, **kwargs)

        return super().to(*args, *kwargs)

    def dist_to(self, dist, *args, **kwargs):
        for key, value in dist.__dict__.items():
            if torch.is_tensor(value):
                dist.__dict__[key] = value.to(*args, **kwargs)
            elif isinstance(value, distributions.Distribution):
                self.dist_to(value, *args, **kwargs)


class PyroSigmoid(nn.Sigmoid, PyroModule):
    pass


class PyroTanh(nn.Tanh, PyroModule):
    pass


class PyroReLU(nn.ReLU, PyroModule):
    pass


class BNNNotSampledError(Exception):
    pass


class PyroMCMCBNN(nn.Sequential, PyroModule):
    def __init__(self, *args, sigma=1.0, step_size=1e-6, num_steps=50):
        super().__init__(*args)

        self.sigma = sigma

        # self.hmc_kernel = HMC(self, step_size=step_size, num_steps=num_steps, jit_compile=True, ignore_jit_warnings=True)
        self.hmc_kernel = NUTS(self, step_size=step_size, full_mass=False, max_tree_depth=6, jit_compile=True, ignore_jit_warnings=True)
        self.mcmc = None

    def forward(self, X, y=None):
        mean = super().forward(X)

        if y is not None:
            with pyro.plate("data", device=X.device):
                obs = pyro.sample("obs", dist.Normal(mean, self.sigma), obs=y)

        return mean

    def sample(self, X, y, num_samples=200, reject=200):
        self.mcmc = MCMC(self.hmc_kernel, num_samples=num_samples, warmup_steps=reject)
        self.mcmc.run(X, y)

    def func_index(self, func, sample_indices, *args, **kwargs):
        if self.mcmc is None:
            raise BNNNotSampledError()

        sample_indices = self.clean_index(sample_indices, self.mcmc.num_samples).to(self._first_sample.device)
        samples = select_samples_by_idx(self.mcmc._samples, sample_indices)

        predictive = Predictive(func, posterior_samples=samples, return_sites=('_RETURN',), parallel=True)
        y = predictive(*args, **kwargs)['_RETURN']

        return y

    @property
    def _first_sample(self):
        return next(iter(self.mcmc._samples.values()))

    def clean_index(self, indices, length):
        def saturate(i):
            return min(max(i, -length), length - 1)

        if isinstance(indices, slice):
            start = saturate(indices.start)
            stop = None if indices.stop is None else saturate(indices.stop)

            indices = list(range(start, stop, indices.step))

        indices = torch.as_tensor(indices)
        if torch.any((indices < -length) | (indices >= length)):
            raise IndexError('sample index out of range')

        # Shift negative to positive
        return (indices + self.mcmc.num_samples) % self.mcmc.num_samples

    def predict_index(self, sample_indices, X):
        return self.func_index(self, sample_indices, X)

    def predict_dist(self, X, num_samples=None):
        if self.mcmc is None:
            raise BNNNotSampledError()

        predictive = Predictive(self, posterior_samples=self.mcmc.get_samples(num_samples), return_sites=('_RETURN',), parallel=True)
        y = predictive(X)['_RETURN']
        return y

    def predict_mean(self, X, num_samples=None):
        y = self.predict_dist(X, num_samples=num_samples)
        return y.mean(0)

    def to(self, *args, **kwargs):
        for module in self:
            module.to(*args, **kwargs)

        return super().to(*args, *kwargs)


def select_samples_by_idx(samples, sample_indices, group_by_chain=False):
    """
    Performs selection from given MCMC samples.

    :param dictionary samples: Samples object to sample from.
    :param IntTensor sample_indices: Indices of samples to return.
    :param bool group_by_chain: Whether to preserve the chain dimension. If True,
        all samples will have num_chains as the size of their leading dimension.
    :return: dictionary of samples keyed by site name.
    """
    if not samples:
        raise ValueError("No samples found from MCMC run.")
    if group_by_chain:
        batch_dim = 1
    else:
        samples = {k: v.reshape((-1,) + v.shape[2:]) for k, v in samples.items()}
        batch_dim = 0
    samples = {k: v.index_select(batch_dim, sample_indices) for k, v in samples.items()}
    return samples



import functools
import random
import time

import numpy as np
import torch


class ResetableModule:
    def reset(self):
        def _reset(m):
            if hasattr(m, 'reset_parameters'):
                m.reset_parameters()
        self.apply(_reset)


def timer(func):
    """Print the runtime of the decorated function"""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value

    return wrapper_timer


def set_random_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)



from typing import Callable, Tuple

import torch
from torch import nn

from bayne.bounds.util import notnull, add_method


def crown(model):
    @torch.no_grad()
    def crown(self: nn.Sequential, lower, upper):
        LBs, UBs = [lower], [upper]

        modules = surrogate_model(self)

        for i in range(1, len(modules)):
            subnetwork = modules[:i]
            lb, ub = subnetwork_crown(subnetwork, LBs, UBs, lower.device)
            LBs.append(lb)
            UBs.append(ub)

            batch_size = LBs[0].size(0)

        alpha, beta = alpha_beta_sequential(modules, LBs, UBs)
        bounds = linear_bounds(modules, alpha, beta, batch_size, lower.device)

        return bounds

    add_method(model, 'crown', crown)

    return model


def surrogate_model(model):
    with notnull(getattr(model, '_pyro_context', None)):
        layers = []

        for module in model:
            with notnull(getattr(module, '_pyro_context', None)):
                if isinstance(module, nn.Linear):
                    layers.append(SurrogateLinear(module.weight, module.bias))
                else:
                    layers.append(module)

        return layers


class SurrogateLinear:
    def __init__(self, weight, bias):
        self.weight = weight
        self.bias = bias
        self.out_features = weight.size(-2)


def subnetwork_crown(model, LBs, UBs, device):
    batch_size = LBs[0].size(0)

    alpha, beta = alpha_beta_sequential(model, LBs, UBs)
    bounds = linear_bounds(model, alpha, beta, batch_size, device)

    return interval_bounds(bounds, LBs[0], UBs[0])


def interval_bounds(bounds, lower, upper):
    (Omega_0, Omega_accumulator), (Gamma_0, Gamma_accumulator) = bounds

    lower, upper = lower.unsqueeze(-2), upper.unsqueeze(-2)

    # We can do this instead of finding the Q-norm, as we only deal with perturbation over a hyperrectangular input,
    # and not a B_p(epsilon) ball
    mid = (lower + upper) / 2
    diff = (upper - lower) / 2

    min_Omega_x = (torch.matmul(Omega_0, mid) - torch.matmul(torch.abs(Omega_0), diff))[..., 0]
    max_Gamma_x = (torch.matmul(Gamma_0, mid) + torch.matmul(torch.abs(Gamma_0), diff))[..., 0]

    return min_Omega_x + Omega_accumulator, max_Gamma_x + Gamma_accumulator


def alpha_beta_sequential(model, LBs, UBs):
    alpha, beta = [], []

    for module, LB, UB in zip(model, LBs, UBs):
        assert torch.all(LB <= UB + 1e-6)

        n = UB <= 0
        p = 0 <= LB
        np = (LB < 0) & (0 < UB)

        if isinstance(module, nn.ReLU):
            a, b = alpha_beta_relu(LB, UB, n, p, np)
        elif isinstance(module, nn.Sigmoid):
            a, b = alpha_beta_sigmoid(LB, UB, n, p, np)
        elif isinstance(module, nn.Tanh):
            a, b = alpha_beta_tanh(LB, UB, n, p, np)
        else:
            a, b = (None, None), (None, None)

        alpha.append(a)
        beta.append(b)

    return alpha, beta


def alpha_beta_relu(LB, UB, n, p, np, adaptive_relu=True):
    alpha_lower_k = torch.zeros_like(LB)
    alpha_upper_k = torch.zeros_like(LB)
    beta_lower_k = torch.zeros_like(LB)
    beta_upper_k = torch.zeros_like(LB)

    alpha_lower_k[n] = 0
    alpha_upper_k[n] = 0
    beta_lower_k[n] = 0
    beta_upper_k[n] = 0

    alpha_lower_k[p] = 1
    alpha_upper_k[p] = 1
    beta_lower_k[p] = 0
    beta_upper_k[p] = 0

    LB, UB = LB[np], UB[np]

    z = UB / (UB - LB)
    if adaptive_relu:
        a = (UB >= torch.abs(LB)).to(torch.float)
    else:
        a = z

    alpha_lower_k[np] = a
    alpha_upper_k[np] = z
    beta_lower_k[np] = 0
    beta_upper_k[np] = -LB * z

    return (alpha_lower_k, alpha_upper_k), (beta_lower_k, beta_upper_k)


def alpha_beta_sigmoid(LB, UB, n, p, np):
    def derivative(d):
        return torch.sigmoid(d) * (1 - torch.sigmoid(d))

    return alpha_beta_general(LB, UB, n, p, np, torch.sigmoid, derivative)


def alpha_beta_tanh(LB, UB, n, p, np):
    def derivative(d):
        return 1 - torch.tanh(d) ** 2

    return alpha_beta_general(LB, UB, n, p, np, torch.tanh, derivative)


def alpha_beta_general(LB, UB, n, p, np, func, derivative):
    alpha_lower_k = torch.zeros_like(LB)
    alpha_upper_k = torch.zeros_like(LB)
    beta_lower_k = torch.zeros_like(LB)
    beta_upper_k = torch.zeros_like(LB)

    LB_act, UB_act = func(LB), func(UB)
    LB_prime, UB_prime = derivative(LB), derivative(UB)

    d = (LB + UB) * 0.5  # Let d be the midpoint of the two bounds
    d_act = func(d)
    d_prime = derivative(d)

    slope = (UB_act - LB_act) / (UB - LB)

    # Negative regime
    alpha_lower_k[n] = d_prime[n]
    alpha_upper_k[n] = slope[n]
    beta_lower_k[n] = d_act[n] - alpha_lower_k[n] * d[n]
    beta_upper_k[n] = UB_act[n] - alpha_upper_k[n] * UB[n]

    # Positive regime
    alpha_lower_k[p] = slope[p]
    alpha_upper_k[p] = d_prime[p]
    beta_lower_k[p] = LB_act[p] - alpha_lower_k[p] * LB[p]
    beta_upper_k[p] = d_act[p] - alpha_upper_k[p] * d[p]

    #################
    # Crossing zero #
    #################
    # Upper
    UB_prime_at_LB = UB_prime * (LB - UB) + UB_act
    direct_upper = np & (UB_prime_at_LB <= 0)
    implicit_upper = np & (UB_prime_at_LB > 0)

    alpha_upper_k[direct_upper] = slope[direct_upper]

    def f_upper(d):
        a_slope = (func(d) - func(LB[implicit_upper])) / (d - LB[implicit_upper])
        a_derivative = derivative(d)
        return a_slope - a_derivative

    d_upper = bisection(torch.zeros_like(UB[implicit_upper]), UB[implicit_upper], f_upper)

    alpha_upper_k[implicit_upper] = derivative(d_upper[0])
    beta_upper_k[np] = LB_act[np] - alpha_upper_k[np] * LB[np]

    # Lower
    LB_prime_at_UB = LB_prime * (UB - LB) + LB_act
    direct_lower = np & (LB_prime_at_UB >= 0)
    implicit_lower = np & (LB_prime_at_UB < 0)

    alpha_lower_k[direct_lower] = slope[direct_lower]

    def f_lower(d):
        a_slope = (func(UB[implicit_lower]) - func(d)) / (UB[implicit_lower] - d)
        a_derivative = derivative(d)
        return a_derivative - a_slope

    d_lower = bisection(LB[implicit_lower], torch.zeros_like(LB[implicit_lower]), f_lower)

    alpha_lower_k[implicit_lower] = derivative(d_lower[1])
    beta_lower_k[np] = UB_act[np] - alpha_lower_k[np] * UB[np]

    return (alpha_lower_k, alpha_upper_k), (beta_lower_k, beta_upper_k)


def bisection(l: torch.Tensor, h: torch.Tensor, f: Callable[[torch.Tensor], torch.Tensor], num_iter=20) -> Tuple[torch.Tensor, torch.Tensor]:
    midpoint = (l + h) / 2

    for _ in range(num_iter):
        y = f(midpoint)

        l[y <= 0] = midpoint[y <= 0]
        h[y > 0] = midpoint[y > 0]

        midpoint = (l + h) / 2

    return l, h


def linear_bounds(model, alpha, beta, batch_size, device):
    model = surrogate_model(model)

    # Compute bounds as two iterations to reduce memory consumption by half
    return oneside_linear_bound(model, alpha, beta, batch_size, device, act_lower), \
           oneside_linear_bound(model, alpha, beta, batch_size, device, act_upper)


def oneside_linear_bound(model, alpha, beta, batch_size, device, act_fn):
    out_size = output_size(model)

    W_tilde = torch.eye(out_size, device=device).unsqueeze(0).expand(batch_size, out_size, out_size)
    acc = 0

    # List is necessary around zip to allow reversing
    for module, (al_k, au_k), (bl_k, bu_k) in reversed(list(zip(model, alpha, beta))):
        with notnull(getattr(module, '_pyro_context', None)):
            if isinstance(module, (SurrogateLinear, nn.Linear)):
                W_tilde, bias = linear(W_tilde, module)
                acc = acc + bias
            elif isinstance(module, (nn.ReLU, nn.Tanh, nn.Sigmoid)):
                W_tilde, bias = act_fn(W_tilde, al_k, au_k, bl_k, bu_k)
                acc = acc + bias
            else:
                raise NotImplemented()

    return W_tilde, acc


def output_size(model):
    for module in reversed(model):
        if isinstance(module, (SurrogateLinear, nn.Linear)):
            return module.out_features


def linear(W_tilde, module):
    bias = module.bias
    weight = module.weight

    if weight.dim() == 2:
        W_tilde_new = torch.matmul(W_tilde, weight)
    else:
        if W_tilde.dim() == 3:
            W_tilde = W_tilde.unsqueeze(0)

        W_tilde_new = torch.matmul(W_tilde, weight.unsqueeze(1))

    if bias is None:
        bias_acc = 0
    elif bias.dim() == 1:
        bias_acc = torch.matmul(W_tilde, bias)
    else:
        bias = bias.view(bias.size(0), 1, bias.size(-1), 1)
        bias_acc = torch.matmul(W_tilde, bias)[..., 0]

    return W_tilde_new, bias_acc


def act_lower(Omega_tilde, al_k, au_k, bl_k, bu_k):
    bias = torch.sum(Omega_tilde * _theta(Omega_tilde, bl_k, bu_k), dim=-1)
    Omega_tilde = Omega_tilde * _omega(Omega_tilde, al_k, au_k)

    return Omega_tilde, bias


def _theta(omega_weight, beta_lower, beta_upper):
    return torch.where(omega_weight < 0, beta_upper.unsqueeze(-2), beta_lower.unsqueeze(-2))


def _omega(omega_weight, alpha_lower, alpha_upper):
    return torch.where(omega_weight < 0, alpha_upper.unsqueeze(-2), alpha_lower.unsqueeze(-2))


def act_upper(Gamma_tilde, al_k, au_k, bl_k, bu_k):
    bias = torch.sum(Gamma_tilde * _delta(Gamma_tilde, bl_k, bu_k), dim=-1)
    Gamma_tilde = Gamma_tilde * _lambda(Gamma_tilde, al_k, au_k)

    return Gamma_tilde, bias


def _delta(gamma_weight, beta_lower, beta_upper):
    return torch.where(gamma_weight < 0, beta_lower.unsqueeze(-2), beta_upper.unsqueeze(-2))


def _lambda(gamma_weight, alpha_lower, alpha_upper):
    return torch.where(gamma_weight < 0, alpha_lower.unsqueeze(-2), alpha_upper.unsqueeze(-2))



import torch
import torch.nn.functional as F

from torch import nn

from bayne.bounds.util import notnull, add_method


def ibp(class_or_obj):
    """
    Propagate a sample interval through a deterministic fully-connected feedforward network.
    This can correspond to a weight sample using MCMC.

    Assumptions:
    - The network is nn.Sequential
    - All layers of the network is nn.Linear or a monotone activation function

    This method does _not_ support weight bounds, only input bounds.

    @misc{wicker2020probabilistic,
      title={Probabilistic Safety for Bayesian Neural Networks},
      author={Matthew Wicker and Luca Laurenti and Andrea Patane and Marta Kwiatkowska},
      year={2020},
      eprint={2004.10281},
      archivePrefix={arXiv},
      primaryClass={cs.LG}
    }
    """

    if isinstance(class_or_obj, nn.Sequential) or (isinstance(class_or_obj, type) and issubclass(class_or_obj, nn.Sequential)):
        return ibp_sequential(class_or_obj)
    elif isinstance(class_or_obj, nn.Linear) or (isinstance(class_or_obj, type) and issubclass(class_or_obj, nn.Linear)):
        return ibp_linear(class_or_obj)
    else:
        return ibp_activation(class_or_obj)


def ibp_sequential(class_or_obj):
    @torch.no_grad()
    def _ibp(self: nn.Sequential, lower: torch.Tensor, upper: torch.Tensor, pre: bool = False):
        with notnull(getattr(self, '_pyro_context', None)):
            LBs, UBs = [], []

            for i, module in enumerate(self):
                if not hasattr(module, 'ibp'):
                    # Decorator also adds the method inplace.
                    ibp(module)

                if pre:
                    LBs.append(lower)
                    UBs.append(upper)

                lower, upper = module.ibp(lower, upper)

            if pre:
                return LBs, UBs
            else:
                return lower, upper

    add_method(class_or_obj, 'ibp', _ibp)
    return class_or_obj


def ibp_linear(class_or_obj):
    @torch.no_grad()
    def ibp(self: nn.Linear, lower: torch.Tensor, upper: torch.Tensor):
        with notnull(getattr(self, '_pyro_context', None)):
            mid = (lower + upper) / 2
            diff = (upper - lower) / 2

            weight = self.weight
            abs_weight = torch.abs(weight)
            bias = self.bias

            if weight.dim() == 2:
                w_mid = F.linear(mid, weight)
                w_diff = F.linear(diff, abs_weight)
            else:
                w_mid = torch.matmul(weight.unsqueeze(1), mid.unsqueeze(-1))[..., 0]
                w_diff = torch.matmul(abs_weight.unsqueeze(1), diff.unsqueeze(-1))[..., 0]

            lower = w_mid - w_diff
            upper = w_mid + w_diff

            if bias is not None:
                if bias.dim() == 2:
                    bias = bias.unsqueeze(-2)

                lower += bias
                upper += bias

            return lower, upper

    add_method(class_or_obj, 'ibp', ibp)
    return class_or_obj


def ibp_activation(class_or_obj):
    @torch.no_grad()
    def ibp(self: nn.Module, lower: torch.Tensor, upper: torch.Tensor):
        with notnull(getattr(self, '_pyro_context', None)):
            return self(lower), self(upper)

    add_method(class_or_obj, 'ibp', ibp)
    return class_or_obj



import types
from contextlib import contextmanager


@contextmanager
def optional(condition, context_manager):
    if condition:
        with context_manager:
            yield
    else:
        yield


@contextmanager
def notnull(context_manager):
    if context_manager is not None:
        with context_manager:
            yield
    else:
        yield


def add_method(class_or_obj, name, func):
    if not isinstance(class_or_obj, type):
        func = types.MethodType(func, class_or_obj)

    setattr(class_or_obj, name, func)
